!
!   Jose Paulo Bonatti - DMD/CPTEC - May 2011
!
MODULE Utils

  USE InputParameters, ONLY: r8, Grav, Rd, Cp, Gamma, GEps, Lat0

  IMPLICIT NONE

  PRIVATE

  PUBLIC :: NewSigma, NewPs, MoistVertExtrap


CONTAINS


SUBROUTINE NewSigma (KmaxOut, KmaxOutp, &
                     DelSigmaOut, SigInterOut, SigLayerOut)

  ! Read in Delta Sigma Values and
  ! Compute Sigma Interface and Layer Values

  IMPLICIT NONE

  INTEGER, INTENT (IN) :: KmaxOut, KmaxOutp

  REAL (KIND=r8), INTENT (IN) :: DelSigmaOut(KmaxOut)

  REAL (KIND=r8), INTENT (OUT) :: SigInterOut(KmaxOutp), &
                                  SigLayerOut(KmaxOut)

  INTEGER :: k

  REAL (KIND=r8) :: RdoCp, CpoRd, RdoCp1

  ! Compute New Sigma Interface Values
  SigInterOut(1)=1.0_r8
  DO k=1,KmaxOut
     SigInterOut(k+1)=SigInterOut(k)-DelSigmaOut(k)
  END DO
  SigInterOut(KmaxOutp)=0.0_r8

  ! Compute New Sigma Layer Values
  RdoCp=Rd/Cp
  CpoRd=Cp/Rd
  RdoCp1=RdoCp+1.0_r8
  DO k=1,KmaxOut
     SigLayerOut(k)=((SigInterOut(k)**RdoCp1-SigInterOut(k+1)**RdoCp1)/ &
                    (RdoCp1*(SigInterOut(k)-SigInterOut(k+1))))**CpoRd
  END DO

END SUBROUTINE NewSigma


SUBROUTINE NewPs (ImaxOut, JmaxOut, KmaxInp, &
                  gTopoInp, gTopoOut, gPsfcInp, &
                  gTvirInp, gPresInp, gPsfcOut, gLnPsOut)

  ! Computes a New Surface Pressure given a New Orography.
  ! The New Pressure is computed assuming a Hydrostatic Balance
  ! and a Constant Temperature Lapse Rate.  Below Ground, the
  ! Lapse Rate is assumed to be -6.5 K/km.

  IMPLICIT NONE

  INTEGER, INTENT (IN) :: ImaxOut, JmaxOut, KmaxInp

  REAL (KIND=r8), INTENT (IN) :: gTopoInp (ImaxOut,JmaxOut) , &
                                 gTopoOut (ImaxOut,JmaxOut) , &
                                 gPsfcInp (ImaxOut,JmaxOut) , &
                                 gTvirInp (ImaxOut,JmaxOut,KmaxInp) , &
                                 gPresInp (ImaxOut,JmaxOut,KmaxInp)

  REAL (KIND=r8), INTENT (OUT) :: gPsfcOut (ImaxOut,JmaxOut) , &
                                  gLnPsOut (ImaxOut,JmaxOut)

  INTEGER :: ls, k, i, j

  REAL (KIND=r8) :: goRd, GammaLR

  REAL (KIND=r8), DIMENSION (ImaxOut,JmaxOut) :: Zu, TopoRef
  CHARACTER(LEN=4):: OptNewPs='pslc'

  goRd=Grav/Rd

  IF(TRIM(OptNewPs)=='psnm')THEN
      ! Computes a New Surface Pressure given a New Orography.
      ! If the pressure to be corrected is the sea level pressure (psnm).
      ! The reference topography is the height referring to sea level
      ! TopoRef=0.0
     TopoRef=0.0_r8
  ELSE IF(TRIM(OptNewPs)=='pslc') THEN
      ! Computes a New Surface Pressure given a New Orography.
      ! If the pressure to be corrected is the surface pressure. 
      ! It already contains topography information. Therefore, 
      ! a topography of reference to the surface pressure of the 
      ! input data is required.
      ! TopoRef=gTopoInp
     TopoRef=gTopoInp 
  END IF

  ! Compute Surface Pressure Below the Original Ground

  ls=0
  k=1
  GammaLR=Gamma
  DO j=1,JmaxOut
     DO i=1,ImaxOut
        Zu(i,j)=TopoRef(i,j)-gTvirInp(i,j,k)/GammaLR* &
                ((gPsfcInp(i,j)/gPresInp(i,j,k))**(-GammaLR/goRd)-1.0_r8)
        IF (gTopoOut(i,j) <= Zu(i,j)) THEN
           IF (ABS(GammaLR) > GEps) THEN
              gPsfcOut(i,j)=gPresInp(i,j,k)*(1.0_r8+GammaLR/gTvirInp(i,j,k)* &
                            (gTopoOut(i,j)-Zu(i,j)))**(-goRd/GammaLR)
           ELSE
              gPsfcOut(i,j)=gPresInp(i,j,k)*EXP(-goRd/gTvirInp(i,j,k)* &
                            (gTopoOut(i,j)-Zu(i,j)))
           END IF
        ELSE
           gPsfcOut(i,j)=0.0_r8
           ls=ls+1
        END IF
     END DO
  END DO

  ! Compute Surface Pressure Above the Original Ground
  DO k=2,KmaxInp
     IF (ls > 0) THEN
        DO j=1,JmaxOut
           DO i=1,ImaxOut
              IF (gPsfcOut(i,j) == 0.0_r8) THEN
                 GammaLR=-goRd*LOG(gTvirInp(i,j,k-1)/gTvirInp(i,j,k))/ &
                         LOG(gPresInp(i,j,k-1)/gPresInp(i,j,k))
                 IF (ABS(GammaLR) > GEps) THEN
                    Zu(i,j)=Zu(i,j)-gTvirInp(i,j,k)/GammaLR* &
                            ((gPresInp(i,j,k-1)/gPresInp(i,j,k))** &
                            (-GammaLR/goRd)-1.0_r8)
                 ELSE
                    Zu(i,j)=Zu(i,j)+gTvirInp(i,j,k)/ &
                            goRd*LOG(gPresInp(i,j,k-1)/gPresInp(i,j,k))
                 END IF
                 IF (gTopoOut(i,j) <= Zu(i,j)) THEN
                    IF (ABS(GammaLR) > GEps) THEN
                       gPsfcOut(i,j)=gPresInp(i,j,k)*(1.0_r8+GammaLR/gTvirInp(i,j,k)* &
                                     (gTopoOut(i,j)-Zu(i,j)))**(-goRd/GammaLR)
                    ELSE
                       gPsfcOut(i,j)=gPresInp(i,j,k)* &
                                     EXP(-goRd/gTvirInp(i,j,k)*(gTopoOut(i,j)-Zu(i,j)))
                    END IF
                    ls=ls-1
                 END IF
              END IF
           END DO
        END DO
     END IF
  END DO

  ! Compute Surface Pressure Over the Top
  IF (ls > 0) THEN
     k=KmaxInp
     GammaLR=0.0_r8
     DO j=1,JmaxOut
        DO i=1,ImaxOut
           IF (gPsfcOut(i,j) == 0.0_r8) THEN
              gPsfcOut(i,j)=gPresInp(i,j,k)* &
                            EXP(-goRd/gTvirInp(i,j,k)*(gTopoOut(i,j)-Zu(i,j)))
           END IF
        END DO
     END DO
  END IF

  ! ln(ps) in cBar (from mBar)
  gLnPsOut=LOG(gPsfcOut/10.0_r8)

END SUBROUTINE NewPs


SUBROUTINE MoistVertExtrap (ImaxInp, JmaxInp, KmaxInp, KmaxInq, &
                            PresInp, TempInp, SpHuInp)

  ! This Code Extrapolates Moisture Up Into Dry Data Layers
  ! Using Exponential Decrease Up To a Specified Value of
  ! Specific Humidity at 80 hPa, then Use Lims Profile Above
  ! See Middle Atmosphere Program Handbook for Map, Volume 22, 
  ! September 1986, Edited by James M. Russell

  IMPLICIT NONE

  INTEGER, INTENT (IN) :: ImaxInp, JmaxInp, KmaxInp, KmaxInq

  REAL (KIND=r8), INTENT (IN) :: PresInp(KmaxInp)

  REAL (KIND=r8), INTENT (IN) :: TempInp(ImaxInp,JmaxInp,KmaxInp)

  REAL (KIND=r8), INTENT (INOUT) :: SpHuInp(ImaxInp,JmaxInp,KmaxInp,1)

  INTEGER :: i, j, k, kDry, jDy, kDp

  REAL (KIND=r8) :: SpHu1, SpHu8, RlHu, Lat, Dy, Dp, Pmin, Pmm, &
                    Fac, xLamb, xLnPm, xLnQm, xLnPq, xLnQq

  REAL (KIND=r8) :: SpHuSat(KmaxInp), SpHuOut(KmaxInp)

  ! SpHuStrat Stores Lims (Satellite 1986) Observed H2O Above 80 hPa:
  ! k=1,8 for 80, 70, .., 10 hPa
  ! j=1,19 for Every 10 degree of latitude (Starting at North Pole)

  REAL (KIND=r8), DIMENSION (8,19) :: SpHuStrat = RESHAPE &
  ((/5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, &
     5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, &
     5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, &
     5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, &
     4.80_r8, 4.55_r8, 4.50_r8, 4.60_r8, 4.70_r8, 4.75_r8, 4.75_r8, 4.70_r8, &
     4.00_r8, 3.85_r8, 3.90_r8, 4.00_r8, 4.25_r8, 4.50_r8, 4.50_r8, 4.60_r8, &
     3.00_r8, 2.75_r8, 3.00_r8, 3.30_r8, 3.70_r8, 4.20_r8, 4.30_r8, 4.50_r8, &
     2.10_r8, 2.00_r8, 2.00_r8, 2.50_r8, 3.00_r8, 3.90_r8, 4.00_r8, 4.00_r8, &
     2.00_r8, 2.00_r8, 2.00_r8, 2.30_r8, 2.80_r8, 3.50_r8, 3.70_r8, 3.75_r8, &
     2.00_r8, 2.00_r8, 2.00_r8, 2.30_r8, 2.70_r8, 3.25_r8, 3.45_r8, 3.75_r8, &
     2.00_r8, 2.00_r8, 2.00_r8, 2.30_r8, 2.80_r8, 3.50_r8, 3.70_r8, 3.75_r8, &
     2.10_r8, 2.00_r8, 2.00_r8, 2.50_r8, 3.00_r8, 3.90_r8, 4.00_r8, 4.00_r8, &
     3.00_r8, 2.75_r8, 3.00_r8, 3.30_r8, 3.70_r8, 4.20_r8, 4.30_r8, 4.50_r8, &
     4.00_r8, 3.85_r8, 3.90_r8, 4.00_r8, 4.25_r8, 4.50_r8, 4.50_r8, 4.60_r8, &
     4.80_r8, 4.55_r8, 4.50_r8, 4.60_r8, 4.70_r8, 4.75_r8, 4.75_r8, 4.70_r8, &
     5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, &
     5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, &
     5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, &
     5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8, 5.00_r8/)&
     , (/8,19/))
  Pmin=80.0_r8
  Pmm= 10.0_r8
  xLnPm=LOG(Pmin)

  ! Get Correct Units of Lims Moisture

  DO j=1,19
    DO k=1,8
      SpHuStrat(k,j)=SpHuStrat(k,j)*1.0E-06_r8
    END DO
  END DO

  ! Longitude and Latitude Loops

  DO j=1,JmaxInp
    Lat=Lat0*(1.0_r8-2.0_r8*REAL(j-1,r8)/REAL(JmaxInp-1,r8))
    DO i=1,ImaxInp

      ! Compute Saturation Specific Humidity (Dimensionless)
      ! From Temperature (K) and Pressure (hPa)

      DO k=1,KmaxInp
        Fac=6.1078_r8*EXP(17.2693882_r8*(TempInp(i,j,k)-273.16_r8)/ &
            (TempInp(i,j,k)-35.86_r8))
        SpHuSat(k)=0.622_r8*Fac/(PresInp(k)-0.378_r8*Fac)
      END DO

      ! Limit Moisture in Lowest KmaxInq Layers: 
      ! RlHu <= 1.0 but >= 0.15
      ! The Latter Avoids the Negative Specific Humidity Problem

      DO k=1,KmaxInq
        SpHuOut(k)=SpHuInp(i,j,k,1)
        RlHu=SpHuInp(i,j,k,1)/SpHuSat(k)
        IF (RlHu <= 0.15_r8) SpHuOut(k)=0.15_r8*SpHuSat(k)
        IF (RlHu > 1.00_r8) SpHuOut(k)=SpHuSat(k)
      END DO

      ! Obtain 80 hPa (k=1) and 10 hPa (k=8) Values by Horizontal Interpolation

      Dy=(90.0_r8-Lat)/10.0_r8+1.0_r8
      jDy=MIN(INT(Dy),8)
      Dy=Dy-REAL(jDy,r8)
      SpHu8=SpHuStrat(1,jDy)*(1.0_r8-Dy)+SpHuStrat(1,jDy+1)*Dy
      SpHu1=SpHuStrat(8,jDy)*(1.0_r8-Dy)+SpHuStrat(8,jDy+1)*Dy
      kDry=KmaxInq+1
      xLnQm=LOG(SpHu8)

      ! Extrapolate Moisture to Minimum Value (SpHu8) at Pressure Pmin (80 hPa)
      ! Use Exponential Decrease from Layer KmaxInq
      ! i.e.  SpHu=SpHu(KmaxInq)*(PresInp/PresInp(KmaxInq)) ** xLamb
      ! where xLamb is Computed to Fit SpHu8 at Pmin and SpHu,PresInp at KmaxInq

      xLnPq=LOG(PresInp(KmaxInq))
      xLnQq=LOG(MAX(SpHuOut(KmaxInq),1.0e-12_r8))
      xLamb=(xLnQm-xLnQq)/(xLnPm-xLnPq)

      DO k=kDry,KmaxInp

        IF (PresInp(k) >= Pmin) THEN
          SpHuOut(k)=SpHuOut(KmaxInq)*(PresInp(k)/PresInp(KmaxInq))**xLamb

        ELSE IF (PresInp(k) >= Pmm) THEN

          ! Above 80 hPa so Complete Linear Interpolation from Table.

          Dp=8.0_r8-PresInp(k)/10.0_r8
          kDp=MIN(INT(Dp),8)
          Dp=Dp-REAL(kDp)
          SpHuOut(k)=SpHuStrat(kDp,jDy)*(1.0_r8-Dp)*(1.0_r8-Dy)+ &
                     SpHuStrat(kDp+1,jDy+1)*Dp*Dy+         &
                     SpHuStrat(kDp,jDy+1)*(1.0_r8-Dp)*Dy+     &
                     SpHuStrat(kDp+1,jDy)*Dp*(1.0_r8-Dy)

        ELSE

          ! Above Pmm (10 hPa) Use Constant Value from Table

          SpHuOut(k)=SpHu1

        END IF

        RlHu=SpHuOut(k)/SpHuSat(k)
        IF (RlHu > 1.00_r8) SpHuOut(k)=SpHuSat(k)

      END DO

      ! Store Extrapolated Moisture

      DO k=1,KmaxInp
        SpHuInp(i,j,k,1)=SpHuOut(k)
      END DO
  
    END DO
  END DO

END SUBROUTINE MoistVertExtrap


END MODULE Utils
